package net.bigeon.gclc.process.io;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.nio.charset.StandardCharsets;
import java.util.concurrent.atomic.AtomicBoolean;

import org.junit.Test;

import net.bigeon.gclc.manager.PipedConsoleInput;
import net.bigeon.gclc.manager.StreamConsoleInput;
import net.bigeon.gclc.tools.ConstantString;
import net.bigeon.test.junitmt.ATestRunnable;
import net.bigeon.test.junitmt.FunctionalTestRunnable;
import net.bigeon.test.junitmt.TestFunction;
import net.bigeon.test.junitmt.ThreadTest;

public class ConnectingConsoleInputTest {

    @Test
    public void testConnect() throws IOException {
        final ConnectingConsoleInput in = new ConnectingConsoleInput();
        final PipedConsoleInput pci = new PipedConsoleInput(null);
        assertNull("Inputs should not transmit before connection", in.prompt(1000));
        in.connect(pci);
        pci.type("Some message");
        assertEquals("Messages should be buffered", "Some message", in.prompt());
    }

    @Test
    public void testDisconnect() throws IOException {
        final ConnectingConsoleInput in = new ConnectingConsoleInput();
        final PipedConsoleInput pci = new PipedConsoleInput(null);
        assertNull("Inputs should not transmit before connection", in.prompt(1000));
        in.connect(pci);
        pci.type("Some message");
        assertEquals("Messages should be buffered", "Some message", in.prompt());
        in.disconnect();
        assertNull("Disconnected inputs should not transmit anymore", in.prompt(1000));
    }

    @Test
    public void testIsClosed() throws IOException {
        final ConnectingConsoleInput in = new ConnectingConsoleInput();
        assertFalse("New input should not be closed", in.isClosed());
        final PipedConsoleInput pci = new PipedConsoleInput(null);
        in.connect(pci);
        pci.close();
        assertFalse("Connected input closing should not close connecting", in.isClosed());
        final PipedConsoleInput pci2 = new PipedConsoleInput(null);
        in.connect(pci2);
        in.close();
        assertTrue("Input should close on request", in.isClosed());
        assertFalse("Connected input should not be closed by connecting closure",
                pci2.isClosed());
    }

    @Test
    public void testCoveragePrompt() {
        final ConnectingConsoleInput in = new ConnectingConsoleInput();
        in.setPrompt("test");
        assertEquals("Prompt should be set correctly", "test",
                ((ConstantString) in.getPrompt()).get());
        final ConstantString prompt = new ConstantString("other");
        in.setPrompt(prompt);
        assertEquals("Prompt should be set correctly", prompt, in.getPrompt());
    }

    @Test
    public void testPromptSequence() throws IOException, InterruptedException {
        final ConnectingConsoleInput in = new ConnectingConsoleInput();
        // Unconnected
        final AtomicBoolean ended = new AtomicBoolean(false);
        final TestFunction one = new TestFunction() {

            @Override
            public void apply() throws Exception {
                try {
                    in.prompt("m1", -1);
                    fail("interruption of infinite waiting prompt should cause error");
                } catch (final InterruptedIOException e) {
                    // ok
                }
                try {
                    in.prompt("m2", 5000);
                    fail("interruption of finite waiting prompt should cause error");
                } catch (final InterruptedIOException e) {
                    // ok
                }
                synchronized (ended) {
                    ended.set(true);
                    try {
                        assertNull("Overtime should return null", in.prompt("m3", 200));
                    } catch (final InterruptedIOException e) {
                        fail("Unexpected interruption error in overtime");
                    }
                }
            }
        };
        final ATestRunnable runnable = new FunctionalTestRunnable(one);
        final Thread th = new Thread(runnable);
        final Thread inter = new Thread(new Runnable() {

            @Override
            public void run() {
                while (!ended.get()) {
                    try {
                        th.join(100);
                    } catch (final InterruptedException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                    synchronized (ended) {
                        if (!ended.get()) {
                            in.interruptPrompt();
                        }
                    }
                }
            }
        });
        th.start();
        inter.start();
        ThreadTest.assertRuns(th, runnable);
        inter.join();
        ended.set(false);

        final PipedOutputStream os = new PipedOutputStream();
        final PipedInputStream pis = new PipedInputStream(os);
        in.connect(new StreamConsoleInput(null, pis, StandardCharsets.UTF_8));
        final ATestRunnable runnable2 = new FunctionalTestRunnable(one);
        final Thread th2 = new Thread(runnable2);
        final Thread inter2 = new Thread(new Runnable() {

            @Override
            public void run() {
                while (!ended.get()) {
                    try {
                        th2.join(100);
                    } catch (final InterruptedException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                    synchronized (ended) {
                        if (!ended.get()) {
                            in.interruptPrompt();
                        }
                    }
                }
            }
        });
        th2.start();
        inter2.start();
        ThreadTest.assertRuns(th2, runnable2);
    }
}
